#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <signal.h>
#include <getopt.h>
#include <errno.h>

#include <sys/ioctl.h>
#include <linux/kd.h>
#include <linux/types.h>

#include "datatype.h"
#include "if.h"
#include "config.h"
#include "ledctrl.h"
#include "daemon.h"
#include "utils.h"

const char *NETLED_VERSION_STRING = "NetLED Version 4.0.10";
int         debugging = 0;
int         ttyfd;

void show_help ( void ) {
  fprintf(stderr, "Usage: netled [OPTION]\n");
  fprintf(stderr, "  -C, --config=[file]          use a specific configuration fil\n");
  fprintf(stderr, "                               (default=/etc/netled.conf)\n");
  fprintf(stderr, "  -t, --tty=[tty]              select the terminal you wish to control\n");
  fprintf(stderr, "                               (default=console)\n");
  fprintf(stderr, "  -r, --refresh=[ms]           delay between each poll of the interface\n");
  fprintf(stderr, "                               (default=100)\n");
  fprintf(stderr, "  -R, --rx=[led]               LED to toggle on inbound network traffic\n");
  fprintf(stderr, "                               (default=scroll)\n");
  fprintf(stderr, "  -T, --tx=[led]               LED to toggle on outbound network traffic\n");           
  fprintf(stderr, "                               (default=num)\n");
  fprintf(stderr, "  -i, --interface=[interface]  the network interface to monitor\n");
  fprintf(stderr, "                               (default=eth0)\n");
  fprintf(stderr, "  -d, --daemonize              run NetLED as a daemon\n");
  fprintf(stderr, "  -k, --kill                   kill a running NetLED daemon\n");
  fprintf(stderr, "  -q, --query                  find the PID of the NetLED daemon\n");
  fprintf(stderr, "  -l, --list                   list available interfaces\n");
  fprintf(stderr, "  -h, --help                   display this help and exit\n");
  fprintf(stderr, "  -v, --version                display version info and exit\n");
  fprintf(stderr, "  -D, --debug                  display debugging information\n\n");
}

// main application procedure
int main ( int argc, char **argv ) {
  char tty[15] = "/dev/console";  /* default to the console */
  int waitdelay = 100, i = 0;  
  char interface[16] = "eth0";
  if_throughput_t interdat = { "eth0", 0, 0, 0, 0, 0, 0 };
  int rx_led = LED_SCR;
  int tx_led = LED_NUM;

  /* process command line arguments */
  while ( 1 ) {
    int option_index = 0;
    int c;

    static struct option long_options[] = {
      {"tty", 1, 0, 't'},
      {"interface", 1, 0, 'i'},
      {"list", 0, 0, 'l'},
      {"refresh", 1, 0, 'r'},
      {"help", 0, 0, 'h'},
      {"version", 0, 0, 'v'},
      {"tx", 1, 0, 'T'},
      {"rx", 1, 0, 'R'},
      {"query", 0, 0, 'q'},
      {"daemonize", 0, 0, 'd'},
      {"kill", 0, 0, 'k'},
      {"debug", 0, 0, 'D'},
      {0, 0, 0, 0}
    };
     
    c = getopt_long(argc, argv, "t:i:lr:hvT:R:qdkD", 
        long_options, &option_index);
    
    if(c == -1)
      break;
    
    switch(c) {
      case 'T':                                   // --tx
        if(strcasecmp(optarg, "off") == 0) {
          tx_led = -1;
        } else if(strcasecmp(optarg, "caps") == 0) {
          tx_led = LED_CAP;
        } else if(strcasecmp(optarg, "cap") == 0) {
          tx_led = LED_CAP;
        } else if(strcasecmp(optarg, "num") == 0) {
          tx_led = LED_NUM;
        } else if(strcasecmp(optarg, "scroll") == 0) {
          tx_led = LED_SCR;
        } else if(strcasecmp(optarg, "scr") == 0) {
          tx_led = LED_SCR;
        } else {
          printf("Error: invalid --tx option: %s\n", optarg);
          exit(1);
        }
        break;

      case 'R':                                   // --rx
        if(strcasecmp(optarg, "off") == 0) {
          rx_led = -1;
        } else if(strcasecmp(optarg, "caps") == 0) {
          rx_led = LED_CAP;
        } else if(strcasecmp(optarg, "cap") == 0) {
          rx_led = LED_CAP;
        } else if(strcasecmp(optarg, "num") == 0) {
          rx_led = LED_NUM;
        } else if(strcasecmp(optarg, "scroll") == 0) {
          rx_led = LED_SCR;
        } else if(strcasecmp(optarg, "scr") == 0) {
          rx_led = LED_SCR;
        } else {
          printf("Error: invalid --rx option: %s\n", optarg);
          exit(1);
        }
        break;

      case 't':                                   // --tty
        strcpy(tty, "/dev/");
        strncat(tty, optarg, sizeof(tty)/sizeof(char) - 5);
        break;
        
      case 'i':                                   // --interface
        if ( debugging ) fprintf(stderr, "Attempting to use interface %s.", optarg);
        strncpy ( interface, optarg, sizeof(interface) );
        break;
        
      case 'r':                                   // --refresh
        waitdelay = (int)strtol(optarg, NULL, 10);
        if(errno == ERANGE) {
          printf("Warning: Invalid refresh rate specified, using 100: %s\n", 
              optarg);
          waitdelay = 100;
        } 
        
        if(waitdelay < 10) {
          printf("Warning: Refresh rate is tiny, using 10 instead: %s\n", 
              optarg);
          waitdelay = 10;
        }
        
        if(waitdelay > 10000) {
          printf("Note: Refresh rate is quite large." 
                 "  Is this really what you want?: %s\n", optarg);
        }
        break;
        
      case 'l':                                   // --list
        printf("Available interfaces: ");
        show_interfaces();
        printf("\n");
        exit(0);
        
      case 'v':                                   // --version
        printf("%s\n", NETLED_VERSION_STRING);
        printf("Written by Matthew Bevan <mbevan@marginsoftware.com>\n\n");
        printf("Copyright (c) 2002, Margin Software.  All Rights Reserved.\n");
        printf("This is free software; see the source for copying conditions.  There is NO\n");
        printf("warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n");
        exit(0);

      case 'h':                                   // --help
        show_help();
        exit(0);

      case 'd':
	daemonized = 1;
        break;

      case 'k':
	i = get_pid();
	if ( i > 0 ) {
	  if ( kill ( i, SIGTERM ) != 0 ) {
	    fprintf (stderr, "Error: could not kill process %i.\n", i);
	  }
	} else {
	  fprintf(stderr, "Error: no netled process running.\n");
	}
	exit(0);

      case 'q':
        i = get_pid();
	if ( i > 0 ) {
	  fprintf (stderr, "NetLED process running with PID %i.\n", i);
	} else {
	  fprintf (stderr, "Error: no netled process running.\n");
	}
	exit(0);
        
      case 'D':
        debugging = 1;
        break;
        
      default:
        printf("Error: Unexpected getopt() return value: %c\n", c);
        exit(1);
    }
  }

  // Check to see if the chosen interface is valid (i.e. up).  
  if ( !is_interface_valid(interface) ) {
    printf("Error: Unknown interface %s.\nValid interfaces: ", interface);
    show_interfaces();
    printf("\n");
    exit(1);
  }
  
  // Open the terminal.
  if ( (ttyfd = open(tty,O_RDWR) ) < 0 ) {
      printf("Error: Unable to open tty: %s\n", tty);
      exit(1);
  }

  if ( daemonized != 0 ) daemonize ( );

  // organize exit and signal routines...
  atexit(reset_keyboard);
  signal(SIGINT, exit);
  
  strcpy (interdat.if_name, interface);
  
  while ( 1 ) {
    unsigned char leds;
    
    if (!get_if_throughput(&interdat)) {
      unsigned char leds = 0x00; // clear the LEDs.
      
      // read from the real status flags...
      
          
      if ( interdat.chg_bytes_in > 0 ) {
        if ( debugging ) fprintf ( stderr, "RX %X ON \t", rx_led );
        leds |= rx_led;
      } else {
        if ( debugging ) fprintf ( stderr, "RX %X OFF\t", rx_led );
      }

      if ( interdat.chg_bytes_out > 0 ) {
        if ( debugging ) fprintf ( stderr, "TX %X ON \n", tx_led );
        leds |= tx_led;
      } else {
        if ( debugging ) fprintf ( stderr, "TX %X OFF\n", tx_led );
      }
      
      toggleleds ( ttyfd, leds );
    }

    usleep(waitdelay*1000);
  }
  
  return 0;
}
